////////////////////////////////////////////////////////////////////////
// Class:       Deconvolution
// Plugin Type: producer (art v3_05_01)
// File:        Deconvolution_module.cc
//
// Generated at Mon Jan  4 14:57:41 2021 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
//
// About: do deconvolution on waveform from recob::Wire
//        i.e., waveform roi
//
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "canvas/Persistency/Common/Assns.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib_except/exception.h"
#include "cetlib/search_path.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"

// larsoft includes
#include "lardataobj/RecoBase/Wire.h"
#include "lardata/ArtDataHelper/WireCreator.h"
#include "lardata/Utilities/LArFFT.h"
#include "lardata/Utilities/AssociationUtil.h"
#include "lardata/Utilities/SignalShaping.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcore/Geometry/Geometry.h"
#include "larcoreobj/SimpleTypesAndConstants/RawTypes.h" // raw::ChannelID_t

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

// ArgoNeuT includes
#include "T962Utilities/SignalShapingServiceT962.h"

#include <memory>

// ROOT includes
#include "TFile.h"
#include "TH1D.h"
#include "TMath.h"

class Deconvolution;


class Deconvolution : public art::EDProducer {
public:
  explicit Deconvolution(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  Deconvolution(Deconvolution const&) = delete;
  Deconvolution(Deconvolution&&) = delete;
  Deconvolution& operator=(Deconvolution const&) = delete;
  Deconvolution& operator=(Deconvolution&&) = delete;

  // Required functions.
  void produce(art::Event& e) override;

private:
  // fhicl parameters
  art::InputTag fWireProducerLabel;
  std::vector<int> fBaseSampleBins;
  bool fSaveSpaceWithROI;
  bool fDodQdxCalib; // true: will apply wire-by-wire calibration

  // function
  void SubtractBaseline(std::vector<float>& holder, int baseSampleBins);

  int fNticks;
  std::map<unsigned int, float> fdQdxCalib; // Map to do wire-by-wire calibration, key is channel number, content is correction factor
};


Deconvolution::Deconvolution(fhicl::ParameterSet const& p)
  : EDProducer{p}  // ,
  // More initializers here.
{
  fWireProducerLabel      = p.get<art::InputTag>("WireProducerLabel", "");
  fBaseSampleBins         = p.get<std::vector<int>>("BaseSampleBins");
  fSaveSpaceWithROI       = p.get<bool>("SaveSpaceWithROI", false);
  fDodQdxCalib            = p.get<bool>("DodQdxCalib", false);
  
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob();
  fNticks = detProp.ReadOutWindowSize(); // number of clock ticks per event

  cet::search_path sp("FW_SEARCH_PATH");
  if (fDodQdxCalib) {
    std::string fileName = p.get< std::string >("dQdxCalibFileName");
    std::string fullname;
    sp.find_file(fileName, fullname);

    if (fullname.empty()) {
      throw cet::exception("Deconvolution") << "dQdxCalib File not found!";
    }
    else {
      mf::LogInfo("Deconvolution") << "Applying wire-by-wire calibration using file " << fileName;
    }

    std::ifstream inFile(fullname, std::ios::in);
    std::string line;

    while (std::getline(inFile,line)) {
      unsigned int channel;
      float        constant;
      std::stringstream linestream(line);
      linestream >> channel >> constant;
      fdQdxCalib[channel] = constant;
      if (channel%100 == 0) std::cout << "Channel " << channel << " correction factor " << fdQdxCalib[channel] << std::endl;
    }
  }
  // Call appropriate produces<>() functions here.
  // Call appropriate consumes<>() for any products to be retrieved by this module.
  produces< std::vector<recob::Wire> >();
}


void Deconvolution::SubtractBaseline(std::vector<float>& holder, int baseSampleBins)
{
  // subtract baseline using linear interpolation between regions defined
  // by the dataSize and baseSampleBins
  if(baseSampleBins <= 0) return;
  constexpr double fBaseVarCut = 20;
  // number of points to characterize the baseline
  unsigned short nBasePts = holder.size() / baseSampleBins;
  // the baseline offset vector
  std::vector<double> base(nBasePts, 0.);
  // find the average value in each region, using values that are
  // similar
  double fbins = baseSampleBins;
  unsigned short nfilld = 0;
  for(unsigned short ii = 0; ii < nBasePts; ++ii) {
    unsigned short loBin = ii * baseSampleBins;
    unsigned short hiBin = loBin + baseSampleBins;
    double ave = 0.;
    double sum = 0.;
    for(unsigned short bin = loBin; bin < hiBin; ++bin) {
      ave += holder[bin];
      sum += holder[bin] * holder[bin];
    } // bin
    ave = ave / fbins;
    double var = (sum - fbins * ave * ave) / (fbins - 1.);
    // Set the baseline for this region if the variance is small
    if(var < fBaseVarCut) {
      base[ii] = ave;
      ++nfilld;
    }
  } // ii

  // interpolate and subtract
  // find the first non-zero sample
  unsigned short thisRegion = 0;
  unsigned short nextRegion = 1;
  for(nextRegion = 1; nextRegion < base.size(); ++nextRegion) if(base[nextRegion] != 0) break;
  if(nextRegion == base.size()) return;
  double nBins = (nextRegion - thisRegion) * baseSampleBins;
  double slp = (base[nextRegion] - base[thisRegion]) / nBins;
  for(unsigned short bin = 0; bin < holder.size(); ++bin) {
    // determine which region we are in
    short region = bin / baseSampleBins;
    if(region == nextRegion) {
      // update the slope for the next region
      unsigned short nnr = 0;
      for(nnr = nextRegion + 1; nnr < base.size(); ++nnr) if(base[nnr] != 0) break;
      if(nnr == base.size()) nnr = base.size() - 1;
      unsigned short previousRegion = thisRegion;
      thisRegion = nextRegion;
      nextRegion = nnr;
      if(thisRegion == nextRegion) {
        for (unsigned short kk = thisRegion * baseSampleBins; kk < thisRegion * baseSampleBins + baseSampleBins; ++ kk) {
          double bs = base[previousRegion] + (double)(kk - previousRegion * baseSampleBins) * slp;
          holder[kk] -= bs;
        }
        return;
      }
      nBins = (nextRegion - thisRegion) * baseSampleBins;
      slp = (base[nextRegion] - base[thisRegion]) / nBins;
    }
    double bs = base[thisRegion] + (double)(bin - thisRegion * baseSampleBins) * slp;

    holder[bin] -= bs;
  } // bin
} // SubtractBaseline


void Deconvolution::produce(art::Event& e)
{
  art::ServiceHandle<util::SignalShapingServiceT962> sss;
  art::ServiceHandle<util::LArFFT> fFFT;
  int transformSize = fFFT->FFTSize();

  // get wire
  art::Handle<std::vector<recob::Wire>> wireListHandle;
  std::vector<art::Ptr<recob::Wire>> wirelist;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wirelist, wireListHandle);
  }

  // make a collection of Wires and an association set
  std::unique_ptr<std::vector<recob::Wire> > wirecol(new std::vector<recob::Wire>);

  // loop over wires
  for (unsigned int ich=0; ich < wirelist.size(); ++ich) {
    const auto & wire = wirelist[ich];
    const auto & signal = wire->Signal();

    int view = wirelist[ich]->View();
    int channel = wirelist[ich]->Channel();

    std::vector<float> adcvec(fNticks);

    for (size_t j=0; j<adcvec.size(); ++j) {
      adcvec[j] = signal[j];
    }

    adcvec.resize(transformSize); // 2048->4096, new elements are initialized as 0 here.
    sss->Deconvolute(channel, adcvec);
    adcvec.resize(fNticks); // 4096->2048, throw away the non-physical section
    SubtractBaseline(adcvec, fBaseSampleBins[view]);

    // apply wire-by-wire calibration
    if (fDodQdxCalib) {
      if (fdQdxCalib.find(channel) != fdQdxCalib.end()) {
        float constant = fdQdxCalib[channel];
        for (int j=0; j < fNticks; j++) {
          adcvec[j] *= constant;
        }
      }
    }

    std::vector<bool> inroi(fNticks, false);

    const recob::Wire::RegionsOfInterest_t& signalROI = wire->SignalROI();

    if (fSaveSpaceWithROI && signalROI.get_ranges().empty()) continue;

    for (const auto& range : signalROI.get_ranges()) {
      // ROI range: [roiFirstBinTick, roiLastBinTick)
      raw::TDCtick_t roiFirstBinTick = range.begin_index();
      raw::TDCtick_t roiLastBinTick = range.end_index();

      for (int j = roiFirstBinTick; j < roiLastBinTick; j++) {
        inroi[j] = true;
      }
    }

    recob::Wire::RegionsOfInterest_t rois(fNticks);

    std::vector<float> sigs;
    int lastsignaltick = -1;
    int roistart = -1;

    for (int j = 0; j < fNticks; ++j) {
      if (inroi[j]) {
        if (sigs.empty()) {
          sigs.push_back(adcvec[j]);
          lastsignaltick = j;
          roistart = j;
        }
        else { 
          if (j != lastsignaltick + 1) {
            rois.add_range(roistart, std::move(sigs));
            sigs.clear();
            sigs.push_back(adcvec[j]);
            lastsignaltick = j;
            roistart = j;
          }
          else {
            sigs.push_back(adcvec[j]);
            lastsignaltick = j;
          }
        }
      }
    }
    if (!sigs.empty()) { rois.add_range(roistart, std::move(sigs)); }

    wirecol->emplace_back(recob::Wire(rois, wirelist[ich]->Channel(), wirelist[ich]->View())); 

  } // end of loop over ich

  e.put(std::move(wirecol));
}

DEFINE_ART_MODULE(Deconvolution)
