////////////////////////////////////////////////////////////////////////////
// Class:       NoiseCorrelation
// Plugin Type: analyzer (art v3_05_01)
// File:        NoiseCorrelation_module.cc
//
// Generated at Mon Jun 29 11:24:41 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
//
// Author: Wanwei Wu, wwu@fnal.gov
//
// About: analyze the correlation between waveforms on all channels.
//    
//   - Correlation study here is based on an empty data event using RawDigit.
//   - Correlation has a time-dependent feature for data, which might be
//     corresponding to run condition, the time dependince of electronics 
//     behavior and possible effects from MINERvA's test.
//   - The time-dependent effect to the coherent noise removal should be 
//     small. Therefore, we can use a selected empty event to do the coherent
//     noise removal approximately.
//   - Current coherent noise removal is based on the correlation study using
//     Run 806, Event 19508.
////////////////////////////////////////////////////////////////////////////

// art framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"

// larsoft includes
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardata/Utilities/LArFFT.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcore/Geometry/Geometry.h"

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

// ROOT includes
#include "TH1.h"
#include "TH2F.h"
#include "TTree.h"
#include "TFile.h"
#include "TCanvas.h"


using namespace std;


class NoiseCorrelation;


class NoiseCorrelation : public art::EDAnalyzer {
public:
  explicit NoiseCorrelation(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  NoiseCorrelation(NoiseCorrelation const&) = delete;
  NoiseCorrelation(NoiseCorrelation&&) = delete;
  NoiseCorrelation& operator=(NoiseCorrelation const&) = delete;
  NoiseCorrelation& operator=(NoiseCorrelation&&) = delete;

  // Required functions.
  void analyze(art::Event const& e) override;

  void beginJob() override;

private:

  // Declare member data here.

  // fcl parameter configuration
  std::string fRawDigitLabel;
  std::string fRawDigitInstance;


  // event information
  int event;
  int run;
  int subrun;

  // TPC information
  int fNticks;

  TH2F *fCorrelationHist[2];
  float fCorrelation;

  TH1I *fBadChannelHist;
  
};


NoiseCorrelation::NoiseCorrelation(fhicl::ParameterSet const& p)
  : EDAnalyzer{p}  // ,
  // More initializers here.
{
  // Call appropriate consumes<>() for any products to be retrieved by this module.
  fRawDigitLabel       = p.get<std::string>("RawDigitLabel");
  fRawDigitInstance    = p.get<std::string>("RawDigitInstance");

  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob();
  fNticks = detProp.NumberTimeSamples(); // number of clock ticks per event
  std::cout << "number of clock ticks per event: " << fNticks << std::endl;
}


void NoiseCorrelation::beginJob()
{
  art::ServiceHandle<art::TFileService> tfs;
  
  for (int p=0; p<2; p++) {
    fCorrelationHist[p] = tfs->make<TH2F>(Form("fCorrelationHist_plane_%d",p), Form("Correlation (plane %d);Channel number;Channel number;", p), 240, 0, 240, 240, 0, 240);
  }

  fBadChannelHist = tfs->make<TH1I>("BadChannelHist", "Bad Channels; Channel number;", 480, 0, 480);

} // beginJob


void NoiseCorrelation::analyze(art::Event const& e)
{
  // Implementation of required member function here.

  // event info
  run = e.run();
  subrun = e.subRun();
  event = e.id().event();

  std::cout << "run: " << run << "; subrun: " << subrun << "; event:  " << event << std::endl;
  
  // channel status
  lariov::ChannelStatusProvider const& channelStatus
    = art::ServiceHandle<lariov::ChannelStatusService const>()->GetProvider();

  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, fRawDigitInstance, rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }

  //std::cout << "rawdigitlist.size(): " << rawdigitlist.size() << std::endl;

  for (int p=0; p<2; p++) {

    // create a 2D vector to save the correlations for each plane
    std::vector<std::vector<float>> correlationArray(240, std::vector<float>(240,0.));

    // get the ADC vector for each channel
    std::vector<std::vector<float>> adcVec;

    // loop over plane channels
    for (size_t i=0; i<240; ++i) {

      // channel number
      int channel = i + 240*p;
      art::Ptr<raw::RawDigit> rawdigit = rawdigitlist[channel];

      // double check the channel number
      if (channel != int(rawdigit->Channel())) {
        throw cet::exception("NoiseCorrelation") << "Error: Wrong channel number, please check!";
      }

      // save bad channel information: for correlation we include all channels
      if (channelStatus.IsBad(channel)) {
        fBadChannelHist->Fill(channel);
      }

      // number of samples in RawDigit data
      int nSamples =  rawdigit->Samples();
      if (nSamples != fNticks) {
        throw cet::exception("NoiseCorrelation") << "Error: number of samples/ticks not constistent";
      }

      // pedestal
      int pedestal = rawdigit->GetPedestal();

      std::vector<short> rawadc(nSamples); // with pedestal

      raw::Uncompress(rawdigit->ADCs(), rawadc, pedestal, rawdigit->Compression());
      
      // subtract pedestal
      std::vector<float> normaladc(nSamples); // subtract pedestal
      for (int j=0; j<nSamples; j++) {
        normaladc.at(j)=float(rawadc.at(j)-pedestal);
      }

      adcVec.push_back(normaladc);
    } // loop plane channel i

    // look at all possible channel combinations
    for (unsigned int i=0; i<240; i++) {
      for (unsigned int j=0; j<240; j++) {

        fCorrelation = -999.;

        std::vector<float> adcvec1 = adcVec[i];
        std::vector<float> adcvec2 = adcVec[j];

        size_t n1 = adcvec1.size();
        size_t n2 = adcvec2.size();

        if (n1 != n2) {
          throw cet::exception("NoiseCorrelation") << "Error: number of samples/ticks not constistent for channel " << i << " and channel " <<j;
        }

        // correlations using the Pearson Correlation Coefficient
        float sumxy = 0, sumx = 0, sumy = 0, sumx2 = 0, sumy2 = 0;
        for (size_t tick = 0; tick<n1; ++tick) {
          double adc1 = adcvec1[tick];
          double adc2 = adcvec2[tick];
          sumx += adc1;
          sumy += adc2;
          sumxy += adc1*adc2;
          sumx2 += adc1*adc1;
          sumy2 += adc2*adc2;
        }

        double  denom = ((n1*sumx2) - (sumx*sumx)) * ((n1*sumy2) - (sumy*sumy));

        if (n1 > 0 && denom > 0) {
          fCorrelation = ( (n1*sumxy) - (sumx*sumy) ) / ( TMath::Sqrt( denom ) );
          correlationArray[i][j] = fCorrelation;
        }

      } // loop channel j
    } // loop channel i

    for (unsigned int channel1 = 0; channel1 < correlationArray.size(); ++channel1) {
      for (unsigned int channel2 = 0; channel2 < correlationArray.size(); ++channel2) {
        fCorrelationHist[p]->SetBinContent(channel1+1, channel2+1, correlationArray[channel1][channel2]);
      }
    }

  } // loop over plane p

} // analyzer

DEFINE_ART_MODULE(NoiseCorrelation)
